//
//  TiledScrollView.m
//  sc
//
//  Created by kunii on 09/09/03.
//  Copyright 2009 __MyCompanyName__. All rights reserved.
//

#import <QuartzCore/QuartzCore.h>
#import "TiledScrollView.h"
#import "testView.h"

#define DEFAULT_TILE_SIZE 256	//	初期タイルサイズ

@implementation TiledScrollView
@synthesize tileSize;
@synthesize tileContainerView;
@synthesize dataSource;

/*
	annotateTileはデバッグ用
	タイルの領域と番号を表示させる。
*/
#define LABEL_TAG 3

- (void)annotateTile:(UIView *)tile {
    static int totalTiles = 0;
    
    UILabel *label = (UILabel *)[tile viewWithTag:LABEL_TAG];
    if (!label) {  
        totalTiles++;		//	この番号は作成時で固定となる。
        UILabel *label = [[UILabel alloc] initWithFrame:CGRectMake(5, 0, 80, 50)];
        [label setBackgroundColor:[UIColor clearColor]];
        [label setTag:LABEL_TAG];
        [label setTextColor:[UIColor greenColor]];
        [label setShadowColor:[UIColor blackColor]];
        [label setShadowOffset:CGSizeMake(1.0, 1.0)];
        [label setFont:[UIFont boldSystemFontOfSize:40]];
        [label setText:[NSString stringWithFormat:@"%d", totalTiles]];
        [tile addSubview:label];
        [label release];
        [[tile layer] setBorderWidth:2];
        [[tile layer] setBorderColor:[[UIColor greenColor] CGColor]];
    }
    
    [tile bringSubviewToFront:label];
}

/*
	初期化
	使っていないタイルビュー（タイル用UIView）保持用インスタンスの確保や、表示タイル領域変数の初期化
	タップ検出、およびタイルビュー埋め込み用ビューの埋め込みをおこなう。
*/
- (id)initWithFrame:(CGRect)frame {
    if (self = [super initWithFrame:frame]) {
		//	各タイル用UIViewは使い回すので、その管理用にNSMutableSetインスタンスを作成
		reusableTiles = [[NSMutableSet alloc] init];
		//	タップ検出用には、各タイルのUIViewを使わず、全体を囲むUIViewを用意する。ここでは以前用意したtestViewを利用。ただしtestViewはdrawRectを無効にした。
		tileContainerView = [[testView alloc] initWithFrame:CGRectZero];
		tileContainerView.drawSelf = NO;				//	drawRectを無効にする。
		[tileContainerView setBackgroundColor:[UIColor redColor]];
		[self addSubview:tileContainerView];
		[self setTileSize:CGSizeMake(DEFAULT_TILE_SIZE, DEFAULT_TILE_SIZE)];
		//	表示に必要な領域のタイルのインディックスを初期化。最初は空にする。
		firstVisibleRow = firstVisibleColumn = NSIntegerMax;
		lastVisibleRow  = lastVisibleColumn  = NSIntegerMin;
		//	ズームに対する移譲を自分で管理することで、スケール別の画像切り替えを対応する。
		[super setDelegate:self];
    }
    return self;
}

- (void)dealloc {
    [reusableTiles release];
    [tileContainerView release];
    [super dealloc];
}

/*
	現在利用されていないUIViewがあればそれを返す。nilを返す場合もある。
*/
- (UIView *)dequeueReusableTile {
    UIView *tile = [reusableTiles anyObject];		//	どれでもいいから、あるやつを取り出す。
    if (tile) {
		/*
			reusableTilesでremoveObjectをするとreleaseされてしまう。このまま返すと無効なインスタンスを渡してしまうことになる。
			そのため、先にretainをおこなうが、そのまま返すと、これまた呼び出し先のretainで2重にretainがかかることになる。
			だからといってreleaseして返せば、ふりだしに戻ってしまう。そのためautoreleaseで自動解放プール（Autorelease Pools）に登録して
			releaseのタイミングをずらした状態で返す。
		*/
        [[tile retain] autorelease];				//	retainで確保。ただし、このメソッドから返って、自動解放プールの管理外になったらreleaseされるように設定。
        [reusableTiles removeObject:tile];			//	reusableTilesから取り外す。ここでreleaseが呼ばれる。
    }
    return tile;
}

/*
	タイルビューを再設定する。
	
	現在tileContainerViewに埋め込まれているUIViewはすべて取り外しreusableTilesに
	登録した上でlayoutSubviewsを呼び出す。
*/
- (void)reloadData {
    /*
		次のlayoutSubviews呼び出しで、tileContainerViewに埋め込まれているUIViewを
		すべて再設定ため、現在埋め込まれているUIViewはすべて取り外しreusableTilesに登録する。
	*/
	for (UIView *view in [tileContainerView subviews]) {
		[reusableTiles addObject:view];
		[view removeFromSuperview];
	}
	//	いったん、何もUIViewが無い状態に設定。
    firstVisibleRow = firstVisibleColumn = NSIntegerMax;
    lastVisibleRow  = lastVisibleColumn  = NSIntegerMin;
	//	layoutSubviewsを呼び出すよう命令。
    [self setNeedsLayout];
}

/*
	内容部を新しい内容に再設定。
	間接的にlayoutSubviewsが呼び出される。
*/
- (void)reloadDataWithNewContentSize:(CGSize)size {
    [self setContentSize:size];
	//	tileContainerViewの大きさも連動させる
    [tileContainerView setFrame:CGRectMake(0, 0, size.width, size.height)];
    //	タイルビューを再設定する。
    [self reloadData];
}

/*
	スクロール、ズームで呼ばれるメソッドの拡張。
*/
- (void)layoutSubviews {
	[super layoutSubviews];					//	元処理をおこなわせる。
	CGRect visibleBounds = [self bounds];	//	表示矩形
	//	tileContainerViewに埋め込まれたタイル用UIViewのうち表示矩形外はすべてreusableTilesに回収する。
	for (UIView *tile in [tileContainerView subviews]) {
		//	タイル用UIViewがvisibleBoundsと交差しているかどうかを判断するためvisibleBoundsの座標系に変換する。
		CGRect scaledTileFrame = [tileContainerView convertRect:[tile frame] toView:self];
		if (! CGRectIntersectsRect(scaledTileFrame, visibleBounds)) {
			//	交差していないならtileContainerViewから取り外しreusableTilesに移動させる。
			[reusableTiles addObject:tile];
			[tile removeFromSuperview];
		}
	}
	//	その大きさに基づいて、tileContainerViewの内容を表示するにはいくつのタイルが必要か計算する。
	int maxRow = floorf([tileContainerView frame].size.height / tileSize.height);	//	縦表示に必要になる最大値
	int maxCol = floorf([tileContainerView frame].size.width  / tileSize.width);	//	横　〃
	//	左、上、右、下のインディックスを計算する。
	int firstNeededRow = MAX(0, floorf(visibleBounds.origin.y / tileSize.height));
	int firstNeededCol = MAX(0, floorf(visibleBounds.origin.x / tileSize.width));
	int lastNeededRow  = MIN(maxRow, floorf(CGRectGetMaxY(visibleBounds) / tileSize.height));
	int lastNeededCol  = MIN(maxCol, floorf(CGRectGetMaxX(visibleBounds) / tileSize.width));

	/*
		このループは、新しく表示されるエリアに属するタイル用UIViewのループ。
		以前の表示エリア内に存在していれば、タイル用UIViewはすでに存在し内容部の更新は不要。
		存在しなければ新たにdataSourceに内容部を更新したタイル用UIViewを用意してもらう必要がある。
	*/
	for (int row = firstNeededRow; row <= lastNeededRow; row++) {
		for (int col = firstNeededCol; col <= lastNeededCol; col++) {
			BOOL needUpdateContents = (firstVisibleRow > row || firstVisibleColumn > col || 
								  lastVisibleRow  < row || lastVisibleColumn  < col);
			if (needUpdateContents) {
				//	新たにdataSourceに内容部を更新したタイル用UIViewを用意してもらう必要がある。
				UIView *tile = [dataSource tiledScrollView:self tileForRow:row column:col resolution:0];			
				//	tileContainerViewの適切な位置にタイル用UIViewを配置する。
				CGRect frame = CGRectMake([self tileSize].width * col, [self tileSize].height * row, [self tileSize].width, [self tileSize].height);
				[tile setFrame:frame];
				[tileContainerView addSubview:tile];
				//	デバッグ用にタイルの領域と番号を表示させる。
				[self annotateTile:tile];
			}
		}
	}
	//	あたらしい表示エリアの設定。
	firstVisibleRow = firstNeededRow; firstVisibleColumn = firstNeededCol;
	lastVisibleRow  = lastNeededRow;  lastVisibleColumn  = lastNeededCol;            
}

/*
	ズームはしない
*/
- (UIView *)viewForZoomingInScrollView:(UIScrollView *)scrollView {
    return nil;
}

/*
	UIScrollViewDelegateは自分が利用するので、他から設定されないようにする。
*/
- (void)setDelegate:(id)delegate {
    NSLog(@"You can't set the delegate of a TiledZoomableScrollView. It is its own delegate.");
}
@end
